/**************************************************************************\
Copyright (C) 2012 Armin Biere JKU Linz

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
****************************************************************************

CONTENTS OF THIS FILE

This file contains the library code for Cleaneling. More precisely it
implements the concrete 'Solver' subclass of 'AbstractSolver'.  The various
parts of this file start with a 'PART' header (search for 'PART').

In 'PART1' we have includes, define macros and some global utility functions
including those for monitoring memory usage.  The next 'PART2' consists of
template code for a 'vector' and a 'priority_queue' container.

Then the internal data structures of 'Solver' are defined in 'PART3'.  After
the declaration of the 'Solver' class itself in 'PART4' the internal member
functions are implemented in 'PART5'.  Finally in 'PART6' we give the
implementation of the public member functions overwriding their virtual
counter-parts in 'AbstractSolver'.

--------------------------------------------------------------------------*
***** PART1: includes, macros, memory monitoring, utilities ***************
--------------------------------------------------------------------------*/

#include "cleaneling.h"

/*------------------------------------------------------------------------*/

#include <cassert>
#include <climits>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <iomanip>
#include <iostream>

/*------------------------------------------------------------------------*/

// Set operating system dependent configuration options here.

#ifdef __linux__
#define HAVE_GETRUSAGE
#else
// If you happen to have 'getrusage' but your are not on Linux, you can
// still just define 'HAVE_GETRUSAGE' here to support timing satistics.
#endif

/*------------------------------------------------------------------------*/

extern "C" {
#ifdef HAVE_GETRUSAGE
#include <sys/time.h>		// use 'getrusage' to gather timing stats
#include <sys/resource.h>
#endif
#if defined(__linux__)
#include <fpu_control.h>	// set FPU to double precision on Linux
#endif
#ifndef NLGL
#include "lglib.h"		// for 'clause_entailed' check ...
#else
struct LGL;			// delcare struct at least
#endif
};

/*------------------------------------------------------------------------*/
// To enable logging code at compile time use the '-l' option, e.g.
// 'configure.sh -l'.  This is norequired when debugging is included, since
// the '-g' option automatically includes logging code.
//
// If logging code has been compiled in, it still has to be enabled at run
// time using the 'log' option through the 'option' set call, or from the
// command line of the binary using the '-l' command line option
// ('cleaneling -l').  The 'log' option is not legal of course if logging
// code is not compiled in.
//
// This macro can only be used within 'Solver' member functions or where at
// least a references to the 'opts' options and the 'level' of the solver is
// available.
//
#ifndef NLOG
#define LOG(CODE) \
  do { \
    if (!opts.log) break; \
    do { cout << "c @" << level << " " << CODE; } while (0); \
    cout << endl; \
  } while (0)
#define LOGCLAUSE(CLAUSE,CODE) \
  do { \
    if (!opts.log) break; \
    do { \
      cout << "c @" << level << " " << CODE; \
      cout << ' ' << ((CLAUSE)->redundant ? "redundant" : "irredundant"); \
      cout << " clause"; \
      for (const int * p = (CLAUSE)->lits; *p; p++) \
	cout << ' ' << *p; \
      cout << endl; \
    } while (0); \
  } while (0)
#else
#define LOG(...) do { } while (0)
#define LOGCLAUSE(...) do { } while (0)
#endif

/*------------------------------------------------------------------------*/

using namespace std;
using namespace Cleaneling;

/*------------------------------------------------------------------------*/
namespace Cleaneling {
/*------------------------------------------------------------------------*/

// If there is just one solver instance we can monitor space usage
// internally.  This is enabled by 'Memory::monitor', which should be called
// before actually allocating the solver.  If there are multiple solvers
// allocated memory usage is over all, thus a global metric.  So this
// feature is reentrant but this 'static' variables are not thread safe.  In
// order to make it thread safe, access to these static variables should be
// guarded by a mutex or something similar, which is currently not
// implemented.
 
static bool monitor_space;
static size_t maximum_bytes;
static size_t current_bytes;

static void inc_bytes (size_t bytes) {			// not thread safe!
  if (!monitor_space) return;
  current_bytes += bytes;
  if (maximum_bytes < current_bytes) maximum_bytes = current_bytes;
}

static void dec_bytes (size_t bytes) {			// not thread safe!
  if (!monitor_space) return;
  assert (current_bytes >= bytes);
  if (current_bytes >= bytes) current_bytes -= bytes;
}

void Memory::monitor () { monitor_space = true; }	// not thread safe!

size_t Memory::maximum () { assert (monitor_space); return maximum_bytes; }

size_t Memory::current () { assert (monitor_space); return current_bytes; }

/*------------------------------------------------------------------------*/

// Process time = user + system time for this process.  This is kind of
// POSIX but not available on Windows.  For now we do not have alternatives.
//
double seconds () {
  double res = 0;
#ifdef HAVE_GETRUSAGE
  struct rusage u;
  if (getrusage (RUSAGE_SELF, &u)) return 0;
  res += u.ru_utime.tv_sec + 1e-6 * u.ru_utime.tv_usec;
  res += u.ru_stime.tv_sec + 1e-6 * u.ru_stime.tv_usec;
#endif
  return res;
}

/*------------------------------------------------------------------------*/

// If you do not tell the hardware to use double precision all the time, you
// risk diverging behaviour of the solver for different levels of
// optimizations.  This is still not really portable and software floats,
// as in PicoSAT or Ligneling should be preferred instead of 'double'.

static void set_fpu_to_double_precision () {
#ifdef __linux__
  fpu_control_t control;
  _FPU_GETCW (control);
  control &= ~_FPU_EXTENDED;
  control &= ~_FPU_SINGLE;
  control |= _FPU_DOUBLE;
  _FPU_SETCW (control);
#endif
}

static double average (double a, double b) { return b ? a/b : 0; }

static double percent (double a, double b) { return b ? 100*a/b : 0; }

static int sign (int lit) { return lit < 0 ? -1 : 1; }

static void internal_error (const char * msg) {
  cout << "internal error in 'libcleaneling.a': " << msg << endl << flush;
  abort ();
}

static long luby (long i) {
  long res = 0, s = 0, k;

  for (k = 1; !res && k < 64; s++, k++) {
    if (i == (1l << k) - 1)
      res = 1l << (k - 1);
  }

  for (k = 1; !res; k++, s++)
    if ((1l << (k - 1)) <= i && i < (1l << k) - 1)
      res = luby (i - (1l << (k-1)) + 1);

  assert (res > 0);

  return res;
}

void print_line (char ch) {
  cout << "c ";
  for (int i = 0; i < 76; i++) cout << ch;
  cout << endl;
}

/*------------------------------------------------------------------------*/
/**** PART2: template code for 'vector' and 'priority_queue' containers ***/ 
/*------------------------------------------------------------------------*/

// A generic stack implementation very similar to 'std::vector' and we try
// to provide a subset of the interface of the STL standard 'std::vector'
// class.  We can enforce calls to 'realloc', which sometimes avoids copying
// memory.  It also updates memory statistics.  Further simplistice
// replacement of 'std::vector' is way easier to use during debugging.
//
// Note that it does not use copy operators for the element types.  So if
// you rely on those to actually copy elements, this will not work.  Same
// holds for elements with destructors.  Those will not be called!

template<class T> class vector {

    T * start, * top, * allocated;

    void realloc (size_t new_size) {
      size_t count = size ();
      size_t old_size = allocated - start;
      size_t old_bytes = old_size * sizeof (T);
      dec_bytes (old_bytes);
      if (new_size) {
	size_t new_bytes = new_size * sizeof (T);
	start = (T*) ::realloc (start, new_bytes);
	if (!start) internal_error ("out of memory reallocating vector");
	inc_bytes (new_bytes);
	top = start + min (count, new_size);
	allocated = start + new_size;
      } else {
	free (start);
	start = top = allocated = 0;
      }
    }

  public:

    vector () : start (0), top (0), allocated (0) { }

    ~vector () { 
      size_t bytes = (allocated - start) * sizeof *start;
      dec_bytes (bytes); 
      free (start);		// do not call destructors for elements!
    }

    bool empty () const { return start == top; }

    size_t size () const { return top - start; }

    T & back () { assert (start < top); return top[-1]; }
    const T & back () const { assert (start < top); return top[-1]; }

    void push_back (const T & data) { 
      if (top == allocated) realloc (size () ? 2 * size () : 1);
      *top++ = data; 
    }

    T & operator[] (size_t i) { assert (i < size ()); return start[i]; }

    const T & operator[] (const size_t i) const { 
      assert (i < size());
      return start[i];
     }

    void pop_back () { assert (start < top); --top; }

    // This shrinks the allocated space to fit the argument size and is not
    // in the API of the vector class of STL.  It also requires that you
    // kind of call destructors of the elements not used anymore manually.
    //
    void shrink (size_t new_size) { 
      assert (new_size <= size ());		// precondition
      realloc (new_size);
    }

    void shrink () { shrink (size ()); }

    void release () { shrink (0); assert (!start && !top && !allocated); }

    // In contrast to 'shrink' this function exists in the STL and has the
    // same meaning.  However it also does not call destructors of removed
    // elements.  You need to do this manually to.
    //
    void resize (size_t new_size) {
      assert (new_size <= size ());		// precondition
      top = start + new_size;
    }

    void resize (const T * new_top) {
      assert (start <= new_top && new_top <= allocated);
      size_t new_size = new_top - start;
      top = start + new_size;
    }

    // Same comments as for 'resize' apply here.  Be carefull when calling
    // this function on vectors which contain elements with destructors.
    //
    void clear () { top = start; }

    typedef int (*Cmp)(T*, T*);
    typedef int (*QsortCmp)(const void *, const void *);

    void sort (Cmp cmp) {
      QsortCmp qsort_cmp = (QsortCmp) cmp;
      qsort (start, size (), sizeof (T), qsort_cmp);
    }

    typedef T * iterator;

    iterator begin () { return start; }
    iterator end () { return top; }

    typedef const T * const_iterator;

    const_iterator begin () const { return start; }
    const_iterator end () const { return top; }
};

/*------------------------------------------------------------------------*/

// A specialized heap implementation, which stores priorities and allows
// fast updates of priorities.  Since this does not exist in the STL, we are
// using our own version and as consequence also run our own 'vector'
// implementation.
//
// The template parameter 'P' gives the type for the priorities which should
// be comparable with '<' (only 'less' needed) parameter.  Currently we use
// 'double' and 'long'. This container also serves as permanent store of
// priorities for all elements ever addded.
//
// Elements in the heap are idexed by non negative integers of type 'int'.
// Therefore the heap as well as the index to the position in heap tables
// only have 'int' elements.  This might prevent this template from being
// used in a more general context.
//
// In general, I think it is hard, if not impossible, to provide an
// efficient, yet simple to use non-intrusive priority queue with updates.
// Making it intrusive is an option, by adding 'piority' and 'position'
// fields to the elements.  However, then it will become harder to allow
// the same element to be in multiple priority queues.  Further, since we
// want to put variables into these queues, storing references to them
// in the queues requires to fix these references if the variable stack
// moves the addresses of the variables when growing due to new variables
// being added.  This was the way how 'Heap' worked in PrecoSAT.

template<class P> class priority_queue {

  vector<int> heap;		// the actual binary heap

  vector<P> prior;		// priorities of type 'P' comparable with '<'

  vector<int> pos;		// maps non negative integer to heap position
  				// negative position = not pushed on heap

  static int left (int p) { return 2*p + 1; }

  static int right (int p) { return 2*p + 2; }

  static int parent (int c) { assert (c > 0); return (c - 1)/2; }

  bool less (int a, int b) { return prior[a] < prior[b]; }

  // The policy for updating the heap, by bubbling elements 'up and 'down,
  // is to only swap two elements if '<' (through the 'less' wrapper)
  // holds and the elements are in the wrong order.  In particular, if the
  // priority of two elements is the same, they are not swapped.

  void up (int e) { 
    int epos = pos[e];
    while (epos > 0) {
      int ppos = parent (epos);
      int p = heap [ppos];
      if (!less (p, e)) break;
      heap [epos] = p, heap [ppos] = e;
      pos[p] = epos, epos = ppos;
    }
    pos[e] = epos;
  }

  void down (int e) {
    assert (contains (e));
    int epos = pos[e], size = (int) heap.size ();
    for (;;) {
      int cpos = left (epos);
      if (cpos >= size) break;
      int c = heap [cpos], o;
      int opos = right (epos);
      if (!less (e, c)) {
	if (opos >= size) break;
	o = heap [opos];
	if (!less (e, o)) break;
	cpos = opos, c = o;
      } else if (opos < size) {
	o = heap [opos];
	if (!less (o, c)) { cpos = opos; c = o; }
      }
      heap [cpos] = e, heap [epos] = c;
      pos[c] = epos, epos = cpos;
    }
    pos[e] = epos;
  }

  // Only if an elment has been 'import'ed before (actually 'push'ed
  // or 'update'd) it has a definite priority.  Otherwise the priority is
  // unknown.
  //
  bool imported (int e) const { 
    assert (0 <= e);
    return e < (int) pos.size (); 
  }

  // Make sure that this element is mapped to a priority.
  //
  void import (int e) {
    while (!imported (e)) {
      pos.push_back (-1);
      prior.push_back (P());
    }
  }

public:

  bool empty () const { return heap.empty (); }

  bool contains (int e) const {
    if (!imported (e)) return false;
    return pos[abs (e)] >= 0;
  }

  // This container also serves as a map of integer elements to priorities.
  // This map can only be changed by 'update'ing a specific priority of
  // an element.  The domain of the map can not be reduced, e.g. it is not
  // possible to 'unmap' an element.  See also 'update' below.
  //
  const P & priority (int e) const {
    assert (imported (e));		// precondition
    return prior[abs (e)];
  }

  // Top element on the heap = largest element.
  //
  int top () const { return heap[0]; }

  // Push an element to the heap either with the default priority 'P()' or
  // if it was on the heap before use the old priority for heap sorting.
  //
  void push (int e) {
    assert (!contains (e));		// precondition
    import (e);
    pos[e] = (int) heap.size ();
    heap.push_back (e);
    assert (heap[pos[e]] == e);
    up (e);
  }

  // Update priority of an elmenent.  There are no restrictions on whether
  // this has been 'push'ed or 'pop'ed.  If is on the heap its position is
  // updated according to the different to the previous priority.  If it not
  // on the heap or has neven been 'push'ed before this function just stores
  // the new priority for this element.
  //
  // This is the only function to change the priority of an element.
  //
  void update (int e, const P & p) {
    import (e);
    P q = prior[e];
    if (q == p) return;
    prior[e] = p;
    if (pos[e] < 0) return;
    if (p < q) down (e);
    if (q < p) up (e);
  }

  // Pop the given argument from heap.  Its priority is kept as is.
  //
  void pop (int e) {
    assert (contains (e));		// precondition
    int i = pos[e];
    pos[e] = -1;
    int last = heap.back ();
    heap.pop_back ();
    int j = (int) heap.size ();
    if (i == j) return;
    assert (i < j);
    pos[last] = i;
    heap[i] = last;
    up (last);
    down (last);
  }

  void pop () { pop (top ()); }

  // Rescore all priorities by multiplying them with the same factor.
  // This is needed by 'Solver::rescore' and can be implemented faster
  // internally to the priority instead of going through 'update'.  An
  // alternative would be to pop and then push all elements (very slow)
  // or allow access to the heap, at least read access in same order as the
  // elments occur on the heap (a constant iterator).
  //
  void rescore (double factor) {
    for (int i = 0; i < (int) prior.size (); i++)
      prior[i] *= factor;
  }
};

/*------------------------------------------------------------------------*/
/**** PART3: internal data structures for 'Solver' ************************/ 
/*------------------------------------------------------------------------*/

struct Option {
  const char * name;		// option name (= opts.#name)
  const char * description;	// usage information for this option
  int val;			// initially default then current value
  int min, max;			// legal option value range

  // Allows to use 'opts.name' instead of 'opts.name.val' for read access to
  // current value of the option.
  //
  operator int () const { return val; }

  Option () { }
  Option (const char * n, int v, int i, int s, const char * d) :
    name (n), description (d), val (v), min (i), max (s) { }
};

// Fore each 'OPTION(NAME,DEFAULT,MIN,MAX,DESCRIPTION)' line in 'options.h'
// add a 'NAME' member to 'Options' with exactly these fields.  Maybe it is
// possible to do this properly in C++.  However, we want to use the
// first argument 'NAME' of 'OPTION' both as string and as C++ symbol and
// this seems to me impossible without further shell scripts to process or
// generate 'options.h'.

struct Options {
  public:
#undef OPTION
#define OPTION(NAME,...) Option NAME;
#include "options.h"
  Options () {
#undef OPTION
#define OPTION(NAME,DEFAULT,MIN,MAX,DESCRIPTION) \
  NAME = Option (#NAME,DEFAULT,MIN,MAX,DESCRIPTION);
#include "options.h"
  }
};

/*------------------------------------------------------------------------*/

struct Clause {
  int glue;			// LBD = glue value as in Glucose

  bool redundant;		// redundant (so not irredundant clause)
  bool remove;			// remove this clause during 'reduce'
  bool important;               // do not reduce (currently if 'glue <= 2')
  bool forcing;			// acting as reason for a literal
  bool dumped;			// dumped but not collected
  bool satisfied;		// top-level satisfied

  long activity;		// how often hit in conflict analysis

  int allocated;		// we reduce the size of clauses
  				// and still need to know the original
				// size which is stored here

  // 'Inlined' and zero terminated list of literals for this clause.
  // Inlining means that 'lits' is actually of an abitrary size.  A zero
  // literal as sentinel is always added at the end, which avoids to store
  // the size explicitly.

  int lits[2];			// of size 2 to avoid 'clang' warnings
  				// the actual size is 'size () + 1' to
				// include the zero sentinel

  // As exampl, for an empty clause 'lits' is really just of size 1,
  // for a unit clauses 'lits' is of size 2, e.g. the unit and 0,
  // for a binary clause 'lits' is of size 3, etc.

  static int cmp (Clause **, Clause **);  // for 'sort' / 'qsort'

  // The following three functions are used in monitoring memory allocated
  // by clauses and is otherwise not used.
  //
  int size () const;
  static size_t bytes (int sz);

  void print () const;		// in DIMACS format for debugging only
};

/*------------------------------------------------------------------------*/

int Clause::size () const {
  size_t res = 0;
  for (const int * p = lits; *p; p++) res++;
  return res;
}

size_t Clause::bytes (int size) {
  return sizeof (Clause) + (size-1) * sizeof (int);
}

void Clause::print () const {
  for (const int * p = lits; *p; p++) cout << *p << ' ';
  cout << '0' << endl;
}

int Clause::cmp (Clause ** p, Clause ** q) {
  Clause * a = *p, * b = *q;
  int res = a->glue - b->glue;
  if (res) return res;
  if (a->activity < b->activity) return 1;
  if (a->activity > b->activity) return -1;
  return 0;
}

/*------------------------------------------------------------------------*/

struct Watch {			// Watcher of a literal in a clause.

  int blit;			// Blocking literal contained in 'clause'.
				// If this literal is true clause satisfied.

  int binary;			// Flag is non zero if clause is binary.
				// Then clause 'clause' does not have to be
				// accessed in propagation.

  Clause * clause;		// Points to actual clause (even if binary).

  Watch (int l, int b, Clause * c) : blit (l), binary (b), clause (c) { }

  Watch ()  { }			// needed for copying
};

typedef vector<Watch> Watches;	// watches of a literal

/*------------------------------------------------------------------------*/

// Full occurrence lists for dense mode.

struct Occs {
  long count;
  vector<Clause*> clauses;
  Occs () : count (0) { }
  void release () { clauses.release (); count = 0; }
  void dec () { assert (count > 0); count--; }
  void add (Clause * c) { count++; clauses.push_back (c); }
};

/*------------------------------------------------------------------------*/

struct Var {			// Variable of index 1..maxvar

  enum State { FREE = 0, FIXED = 1, ELIMINATED = 2, };

  State state;			// see above

  int level;			// decision level if assigned
  int mark;			// mark field for various traverals

  Clause * reason;		// forcing reason clause if assigned

  Watches watches[2];		// 0=pos and 1=neg watches
  Occs occs[2];			// 0=pos and 1=neg occurrences

  Var () : 
    state (FREE),
    level (INT_MAX), 		// needed for sorting watchers
    mark (0), reason (0)
  { }

  bool free () const { return state == FREE; }

  void release () {
    watches[0].release ();
    watches[1].release ();
    occs[0].release ();
    occs[1].release ();
  }
};

/*------------------------------------------------------------------------*/

struct Frame {			// each decision opens up a new frame
  int decision;			// decision at this decision 'level'
  int level;			// decision 'level' of this frame
  int trail;			// trail height before this 'decision'
  int mark;			// frame pulled into 1st UIP clause?
  				// Also used for calculating LBD = glue
  Frame (int d = 0, int l = 0, int t = 0) : 
    decision (d), level (l), trail (t), mark (0) { }
};

/*------------------------------------------------------------------------*/

struct Stats {

  long restarts, iterations, reductions, simplifications;
  long reported, decisions, conflicts, propagations, steps;

  struct {
    int redundant, irredundant; 	// current
    long collected, reduced; 		// accumulated
    long eliminated;
  } clauses;

  struct { int units; long subsumed, strengthened; } distill;

  struct { int fixed, equivalent, eliminated; } vars;
  struct { long learned, minimized; } lits;

  // Time spent in various routines for profiling purposes.
  //
  struct {
     double analyze, bias, extend, minimize, reduce, restart, search;
     double elim, collect, distill, simplify;
     double solve;
  } time;

  Stats () { memset (this, 0, sizeof *this); }	// clear initially
};

/*------------------------------------------------------------------------*/

struct Limits {
  long restart;			// restart after this conflict
  long max_restart_interval;	// only 'report' restart if increased

  struct {
    long redundant;		// reduce limit on redundant clauses 
    long increment;		// increment of limit at next reduction
    long forcing;		// number of forcing redundant clauses
    long important;		// number of 'glue < = 2' clauses
  } reduce;

  Limits () { memset (this, 0, sizeof *this); }	// clear initially
};

/*------------------------------------------------------------------------*/

struct Timer {
  double * timeptr, started;
  Timer (double * tp, double s) : timeptr (tp), started (s) { }
};

/*------------------------------------------------------------------------*/

enum Simplifier { NOSIMP = 0, TOPSIMP = 1, BLOCK = 2, ELIM = 3, };

/*------------------------------------------------------------------------*/
/**** PART4: declaration of 'Solver' **************************************/ 
/*------------------------------------------------------------------------*/

class Solver : public AbstractSolver {

    int level;			// decision level (top/root level == 0)

    bool dense;			// dense mode (full occurrences) or 
				// sparse mode (only watches)

    Simplifier simplifier;	// during simplification set to true

    bool schedule;		// schedule (push) touched literals

    int maxvar () const;	// maximum variable index used/imported

    void import (int lit);	// make sure all data structures are
    				// ready to use this literal / variable

    vector<Var> vars;		// variables
    vector<signed char> vals;	// variable values separately for fast access
    vector<signed char> phases;	// saved phases of previous assignment
    void bias_phases ();	// compute (initial) phase bias 

    int remaining_vars () const;  // remaining variables for verbose output
    				  // without fixed, eliminated etc.

    Var & var (int lit);	  // get variable of literal
    const Var & var (int) const;

    Watches & watches (int lit);		// get literal watches
    const Watches & watches(int lit) const;

    Occs & occs (int lit);			// get full occurrences
    const Occs & occs(int lit) const;

    Value val (int lit) const;	// literal value ('lit' might be negative)

    priority_queue<double> decisions;

    vector<Frame> control;	// at each decisition a frame is added
    vector<int> trail;		// save assigned literals;
    size_t next;		// position of next literal to propagate

    void assign (int, Clause* c = 0);
    void unassign (int);
    void backtrack (int new_level = 0);	// to new level given as argument

    Clause * empty;		// points to an empty clause if non zero
    vector<int> original;	// save original clauses to check assignment
    vector<int> addedlits;	// save added literals before adding clause
    vector<Clause*> clauses;

    LGL * lgl;			// to check entailment ('./configure.sh -c')
    bool clause_entailed ();	// check clause 'addedlits' entailed
    bool trivial_clause ();	// check if clause 'addedlits' is non trivial
    				// and if not remove duplicated literals

    bool each_variable_occurs_only_once ();	// in clause 'addedlits'

    Clause * new_clause (bool redundant = false, int glue = 0);
    void connect_clause (Clause *);	// connect watches
    void dump_clause (Clause *);	// decrease stats but do not delete
    void delete_clause (Clause *);	// physically delete clause memory

    // add occurrence of 'lit' with blocking literal 'blit'
    void add_watch (int lit, int blit, bool binary, Clause *);

    void touch (int lit);		// update and push 'cands.*'
    void add_occ (int lit, Clause * c);	// add full occurrence of 'lit'
    void dec_occ (int lit);		// decrement full occurrence

    void connect_occs (Clause * c);		// for all literals

    void connect_occs ();	// go to dense mode and connect
    				// occurrences of all irredundant clauses

    vector<int> seen;		// save marked literals as 'seen'
    void unmark (int level=0);	// unmark seen literals (after level)
    void mark (int);		// mark literal and save it on 'seen'
    int marked (int) const;	// return signed mark (-1,0,1)

    Status simplify (long steps);	// bounded inprocessing
    Status search (long conflicts);	// bounded search aka CDCL

    bool propagated ();		// all assignments propagated
    Clause * bcp ();		// boolean contstraint propagation
    void analyze (Clause *);	// analyze conflict
    void assume (int decision); // assume decision / probe
    bool decide ();		// returns false if all assigned

    void minimize_clause ();	// minimize the 1st UIP clause
    bool minimize_lit (int);	// try to remove literal from 1st UIP clause

    void restart ();		// backtrack to the top
    void new_restart_limit ();	// set new restart limit
    bool restarting () const;	// restart limit hit

    void reduce ();		// reduce redudant clauses
    bool reducing ();		// reduce limit hit;

    vector<int> frames;		// seen/marked frames during analyze
    int unmark_frames ();	// unmark all those seen frames
    bool mark_frame (int lit);	// mark this frame in analyze
    bool pull_lit (int lit);	// analyze this single literal

    double score_increment;	// current score increment
    void bump_lit (int lit);	// increase score (inverse of decay)
    void bump_clause (Clause *);// increase activity of clause
    void rescore (void);	// rescore score of all literals

    struct {
      priority_queue<long> elim;	// candidates to 'elim'inate
      priority_queue<long> block;	// candidates to 'block'
    } cands;

    void elim (long steps);		// bounded variable elimination

    bool try_elim (int);		// try BVE on this candidate
    void do_elim (int);			// do BVE on this candidate

    bool try_resolve (Clause * c, int cand, Clause * d);
    void do_resolve (Clause * c, int cand, Clause * d);

    bool contains_eliminated (Clause *) const; // contains eliminated variable
    void dump_eliminated ();		       // redundant clauses

    vector<int> probes;			// probes to try
    void probing (long steps);		// simple failed literal probing

    Clause * ignore;		// ignore this and redundant clauses in bcp
    int size_penalty () const;	// reduce simp steps for large formulas
    size_t distilled;		// last distilled clause index
    void distill (long steps);	// distill units, subsume/strengthen clauses

    void disconnect_clauses ();		// disconnect all clauses
    Clause * reduce_clause (Clause *);	// remove false literals
    void collect_clauses ();		// collect garbage clauses
    void connect_clauses ();		// connect all clauses
    void collect ();			// garbage collection

    vector<int> extension;	// saved clauses to extend assignment
    void push_extension (int);	// push (and log) to extension
    void extend ();		// extend partial to full assignment

    void check_satisfied () const;   // check assignment satisfies original

    bool satisfied (Clause * c) const;
    bool satisfied (vector<int>::const_iterator &) const;

    vector<Timer> timers;
    void start (double *);	// start timer and register time stats
    void stop ();		// stop it, update registered time stats

    void report (char type);	// progress report if verbosity is enabled

    Stats stats;
    Limits limits;
    Options opts;

    void print () const;	// in DIMACS format to 'stdout'
    				// for debugging purposes

    void print_time_for (double time, const char *); // print timing stats

  public:

    // exactly the virtual functions in the interface from 'AbstractSolver'
    // make up the public interface of the concrete 'Solver' class

    Solver ();
    ~Solver ();

    void addlit (int);

    Status solve ();
    Value value (int);

    void option (const char *, int);
    int option (const char *);

    void statistics ();
};

/*------------------------------------------------------------------------*/
}; // closing 'namespace Cleaneling'
/*------------------------------------------------------------------------*/

/*------------------------------------------------------------------------*/
/**** PART5: implementation of member private functions of 'Solver' *******/ 
/*------------------------------------------------------------------------*/

int Solver::maxvar () const { 
  int res = (int) vars.size ();
  if (res) { assert (res > 1); res--; }	// avoid under flow
  return res;
}

Var & Solver::var (int lit) { 		// get variable for this literal
  int idx = abs (lit);
  assert (0 < idx && idx < (int) vars.size ());
  return vars[idx];
}

const Var & Solver::var (int lit) const { 
  int idx = abs (lit);
  assert (0 < idx && idx < (int) vars.size ());
  return vars[idx];
}

// Signed current value, that is, if 'lit' is negative the value of the
// variable is negated to get the value of the literal.  The functions
// 'assign' resp. 'unassigned' are used to set resp. reset the value.
//
Value Solver::val (int lit) const {
  int res = vals [abs (lit)];
  if (lit < 0) res = -res;
  return (Value) res;
}

// Signed mark flag, as in 'val' above, i.e. if 'lit' is negative the
// negation of the mark field of the variable is returned.
//
int Solver::marked (int lit) const {
  int res = (int) var (lit).mark;
  return (lit < 0) ? -res : res;
}

// Mark a variable either with '1' resp. '-1' if 'lit' is pos. resp. neg.
// Marked literals are saved on the 'seen' stack, such they can easily be
// unmarked through 'unmark'.
//
void Solver::mark (int lit) {
  Var & v = var (lit);
  assert (!v.mark);
  v.mark = sign (lit);
  seen.push_back (lit);
}

// Unmark all variables previously marked and saved on 'seen'.
//
void Solver::unmark (int level) {
  assert (level <= (int) seen.size ());
  while (level < (int) seen.size ()) {
    int lit = seen.back ();
    seen.pop_back ();
    Var & v = var (lit);
    assert (v.mark == sign (lit));
    v.mark = 0;
  }
}

/*------------------------------------------------------------------------*/

// Compute an initial phase bias based on the Jeroslov-Wang heuristic.

void Solver::bias_phases () {
  start (&stats.time.bias);
  size_t bytes = 2 * (maxvar() + 1) * sizeof (double);
  double * score = (double*) malloc (bytes);
  inc_bytes (bytes);
  if (!score) internal_error ("out of memory allocating score array");
  memset (score, 0, bytes);
  score += maxvar ();
  vector<Clause*>::const_iterator i;
  for (i = clauses.begin (); i != clauses.end (); i++) {
    Clause * c = *i;
    if (c->redundant || c->satisfied) continue;
    double inc = 1;  // compute 2^-size
    for (int size = c->size (); size; size--) inc /= 2.0;
    for (const int * p = c->lits; *p; p++) score[*p] += inc;
  }
  for (int idx = 1; idx <= maxvar (); idx++)
    phases[idx] = ((score[idx] > score[-idx]) ? 1 : -1);
  dec_bytes (bytes);
  score -= maxvar ();
  free (score);
  stop ();
}

/*------------------------------------------------------------------------*/

// We can use an external SAT solver (currently only Lingeling is
// supported), to check whether learned clauses and resolvents are implied
// by the original formula.  This has to be configured at compile time
// either in debug mode or with './configure.sh -c'.  In any case
// the Lingeling library has to accessible in '../lingeling' and checking
// will only be performed if you set the 'check' option.

bool Solver::clause_entailed () {
#ifndef NLGL
  if (!opts.check) return true;
  vector<int>::const_iterator i;
  for (i = addedlits.begin (); i != addedlits.end (); i++)
    lglassume (lgl, -*i);
  bool res = (lglsat (lgl) == 20);
  bool print = false;
  if (!res) print = true;
#ifndef NLOG
  if (opts.log) print = true;
#endif
  if (print) {
    cout << "c @" << level << " checking entailment of clause";
    for (i = addedlits.begin (); i != addedlits.end (); i++)
      cout << ' ' << *i;
    cout << (res ? " succeeded" : " failed") << endl;
  }
  return res;
#else
  return true;
#endif
}

/*------------------------------------------------------------------------*/

// Used in debugging/checking for checking that 'addedlits' all variables
// occur at most once in 'addedlits'.  For external clauses this ensured by
// 'trivial_clause'
//
bool Solver::each_variable_occurs_only_once () {
  int lit;
  vector<int>::const_iterator p;
  for (p = addedlits.begin (); p != addedlits.end (); p++)
    assert (!marked (lit = *p)), assert (!marked (-lit));
  bool res = true;
  for (p = addedlits.begin (); res && p != addedlits.end (); p++)
    if (marked (lit = *p)) res = false;
    else if (marked (-lit)) res = false;
    else mark (lit);
  unmark ();
  return res;
}

// Check if clause is non trivial and if not remove duplicated literals.
//
bool Solver::trivial_clause () {
  vector<int>::iterator q = addedlits.begin ();
  vector<int>::const_iterator p;
  int lit, m;
  bool res = false;
  for (p = q; !res && p != addedlits.end (); p++) {
    lit = *p;
    assert (lit);
    m = marked (lit);
    if (m < 0) res = true; 
    else if (!m) *q++ = lit, mark (lit);
  }
  addedlits.resize (q - addedlits.begin ());
  unmark ();
  return res;
}

Watches & Solver::watches (int lit) { return var (lit).watches[lit < 0]; }

const Watches &
Solver::watches (int lit) const { return var (lit).watches[lit < 0]; }

// Add watch for 'lit' in 'clause' to watcher list of 'lit'.  The size,
// clause and blocking literal are stored as is.  See description of 'Watch'
// above or 'new_clause' below for more details.
//
void Solver::add_watch (int lit, int blit, bool binary, Clause * clause) {
  watches (lit).push_back (Watch (blit, binary, clause));
}

Occs & Solver::occs (int lit) { return var (lit).occs[lit < 0]; }

const Occs &
Solver::occs (int lit) const { return var (lit).occs[lit < 0]; }

void Solver::touch (int lit) {
  if (lit < 0) lit = -lit;
  long new_priority = occs (lit).count + occs (-lit).count;
  if (!var (lit).free ()) return;
  switch (simplifier) {
    case ELIM:
      cands.elim.update (lit, -new_priority);
      if (schedule && !cands.elim.contains (lit))
	cands.elim.push (lit);
      break;
    default:
      break;
  }
}

void Solver::add_occ (int lit, Clause * c) {
  assert (dense);
  occs (lit).add (c);
  touch (lit);
}

void Solver::connect_occs (Clause * c) {
  assert (dense);
  for (const int * p = c->lits; *p; p++)
    add_occ (*p, c);
}

void Solver::connect_occs () {
  assert (!dense && simplifier == TOPSIMP);
  dense = true;
  Clause * c;
  vector<Clause*>::const_iterator i;
  for (i = clauses.begin (); i != clauses.end (); i++)
    if (!(c = *i)->redundant) connect_occs (c);
}

void Solver::dec_occ (int lit) {
  assert (dense);
  occs (lit).dec ();
  touch (lit);
}

/*------------------------------------------------------------------------*/

// Assign literal 'lit' with 'reason' as reason clause .  This assignment
// marks 'lit' as a decision if 'reason' is zero.  Otherwise 'reason' is
// marked as being used as reason to avoid garbage collection during
// 'collect_clauses' and 'reduce'.  Further save this literal as being
// assigned on the 'trail' for backtracking.
//
void Solver::assign (int lit, Clause * reason) {
  Var & v = var (lit);
  assert (!val (lit));
  int idx = abs (lit);
  Value s = (Value) sign (lit);
  vals[idx] = s;
  if (simplifier == NOSIMP) phases[idx] = s;
  if ((v.reason = reason)) {
    assert (!reason->forcing);
    reason->forcing = true; // mark used for 'reduce' and 'collect_clauses'
    if (reason->redundant) limits.reduce.forcing++;
    LOGCLAUSE (reason, "assign " << lit << " forced by");
  } else LOG ("assign " << lit << " decision");
  if (!(v.level = level)) {
    LOG ("fixed " << lit);
    stats.vars.fixed++;
    if (v.state == Var::ELIMINATED) {
      assert (stats.vars.eliminated > 0);
      stats.vars.eliminated--;
    } else assert (v.state == Var::FREE);
    v.state = Var::FIXED;
  }
  trail.push_back (lit);
}

// Unassign a previously assigned literal.  Add it back to 'decisions'.
//
void Solver::unassign (int lit) {
  assert (level > 0);
  Clause * reason;
  Var & v = var (lit);
  assert (val (lit) == TRUE);
  vals [abs (lit)] = UNASSIGNED;
  assert (v.level == level);
  v.level = INT_MAX;
  if ((reason = v.reason)) {
    assert (reason->forcing);
    reason->forcing = false;
    if (reason->redundant) {
      assert (limits.reduce.forcing > 0);
      limits.reduce.forcing--;
    }
  }
  LOG ("unassign " << lit);
  int idx = abs (lit);
  if (!decisions.contains (idx)) decisions.push (idx);
}

// Backtracking involves popping frames from the control stack, unassigning
// all variables in these frames until the last frame still on the control
// stack has the given new decision level.  For decision level 0 we
// initially pushed an empty frame on the control stack, which here serves
// as sentinel.
//
void Solver::backtrack (int new_level) {
  assert (0 <= new_level && new_level <= level);
  if (new_level == level) return;
  LOG ("starting to backtrack to level " << new_level);
  const Frame * f;
  while ((f = &control.back ())->level > new_level) {
    assert (f->level == level);
    assert (f->trail < (int) trail.size ());
    while (f->trail < (int) trail.size ()) {
      int lit = trail.back ();
      assert (var (lit).level == f->level);
      trail.pop_back ();
      unassign (lit);
    }
    assert (level > 0);
    level--;
    LOG ("pop frame decision " << f->decision << " trail " << f->trail);
    trail.resize (f->trail);
    next = f->trail;
    control.pop_back ();
  }
  LOG ("finished backtracking to level " << new_level);
  assert (new_level == level);
}

/*------------------------------------------------------------------------*/

// Allocate, and copy 'addedlits' of previously added literals
// stored in 'addedlits'.  This routine also backtracks to an apropriate
// decision level in which the clause is not empty given the current
// assignment. If it is a unit it will actually also assign the unit
// literal.  If the clause is empty even at the top-level it will then set
// the global 'empty' pointer to this clause if not already an empty clause
// has been found earlier.  Backtracking is actually back jumping to the
// second largest decision level if the clause contains at least two
// literals.
//
Clause * Solver::new_clause (bool redundant, int glue) {
  size_t size;
  Clause * c;

  assert (!glue || redundant);
  assert (each_variable_occurs_only_once ());

  size = addedlits.size ();

  // Allocate and adjust stats (see comments on 'inlining lits' above).
  {
    size_t bytes = Clause::bytes (size);
    c = (Clause *) malloc (bytes);
    if (!c) internal_error ("out of memory while allocating clause");
    inc_bytes (bytes);
#ifndef __clang_analyzer__
    memset (c, 0, bytes);
    c->glue = glue;
    c->important = (glue <= 2);
    c->redundant = redundant;
    c->activity = stats.conflicts;
    c->allocated = size;
#endif
  }

  // Copy literals from 'addedlits' to the result.
  {
    int * p = c->lits;
    vector<int>::const_iterator i;
    for (i = addedlits.begin (); i != addedlits.end (); i++) *p++ = *i;
    assert (!*p);
  }

  // Update clause statistics and log it.
  //
  clauses.push_back (c);
  if (redundant) {
    stats.clauses.redundant++;
    LOGCLAUSE (c, "glue " << glue << " new");
  } else {
    stats.clauses.irredundant++;
    LOGCLAUSE (c, "new");
  }

#ifndef NLGL
  // Add clause to Lingeling for checking entailment of derived clauses.
  //
  if (opts.check) {
    for (const int * p = c->lits; *p; p++) lgladd (lgl, *p);
    lgladd (lgl, 0);
  }
#endif

  return c;
}

/*------------------------------------------------------------------------*/

void Solver::connect_clause (Clause * c) {

  if (c->satisfied) return; // do not connect top level satisfied clauses

  size_t size = c->size ();

  // Swap two literals with maximum level to the front
  // (unassigned variables have 'infinite' level of 'INT_MAX')
  {
    int lit;
    for (int * p = c->lits; *p && p <= c->lits + 1; p++)
      for (int * q = p + 1; (lit = *q); q++) {
	int cmp = (var (lit).level - var (*p).level);
	if (cmp > 0 || (!cmp && (val (lit) > val (*p))))
	  swap (*p, *q);
      }
  }

  int l0 = c->lits[0], l1 = l0 ? c->lits[1] : 0;

  // Backtrack to the minimum of the levels of these two first literals.
  {
    int new_level = (l0 && l1) ? min (var (l0).level, var (l1).level) : 0;
    if (new_level != INT_MAX) backtrack (new_level);
  }

  // Now watch the first two literals if both are non zero, so the clause is
  // at least a binary clause.  See comments to 'Watch' above for details.
  //
  if (size >= 2) {
    add_watch (l0, l1, (size == 2), c);
    add_watch (l1, l0, (size == 2), c);
    if (dense) connect_occs (c);
  }

  // Finally check whether this clause is a unit clause or even empty.
  // If so assign the unit respectively set 'empty' (if still zero).
  { 
    // TODO can we simplify this further?
    bool ignore = false;
    int lit = 0, other, tmp;
    for (const int * p = c->lits; (other = *p) && p <= c->lits + 1; p++) {
      if ((tmp = val (other)) == TRUE) { ignore = true; break; }
      if (tmp == UNASSIGNED) { if (lit) ignore = true; else lit = other; }
    }
    if (!ignore) {
      if (!lit) {
	assert (!level);
	LOG ("found empty clause");
	if (!empty) empty = c;
      } else assign (lit, c);	// Assign unit forced by this clause.
    }
  }

  // Maintain the number of connected important (currently 'glue <= 2')
  // clauses for delaying 'reduce' in case there are too many of them.
  //
  if (c->redundant && c->important) limits.reduce.important++;
}

/*------------------------------------------------------------------------*/

// Pretend that this clause does not exist anymore.

void Solver::dump_clause (Clause * c) {
  if (c->dumped) return;
  if (c->redundant) {
    assert (stats.clauses.redundant > 0);
    stats.clauses.redundant--;
  } else  {
    assert (stats.clauses.irredundant > 0);
    stats.clauses.irredundant--;
    if (dense)
      for (const int * p = c->lits; *p; p++)
	dec_occ (*p);
  }
  c->dumped = true;
  LOGCLAUSE (c, "dump");
}

/*------------------------------------------------------------------------*/

void Solver::delete_clause (Clause * c) {
  dump_clause (c);
  LOGCLAUSE (c, "delete");
  if (monitor_space)				// not thread safe!
    dec_bytes (Clause::bytes (c->allocated));
  free (c);
}

/*------------------------------------------------------------------------*/

// Start timer by saving current time and pointer to the right statistics.
//
void Solver::start (double * time_ptr) {
  timers.push_back (Timer (time_ptr, seconds ()));
}

// Stop the (current = inner most) timer and update its statistics.
//
void Solver::stop () {
  Timer timer = timers.back ();
  timers.pop_back ();
  double delta = seconds () - timer.started;
  if (delta < 0) delta = 0;
  *timer.timeptr += delta;
}

// Imports the given literal and makes sure that all data structures, in
// particularly those involved to represent variables are set up to use this
// literal (and its negation).
//
void Solver::import (int lit) {
  int idx = abs (lit), new_idx;
  assert (lit);
  while (idx >= (new_idx = (int) vars.size ())) {
    vars.push_back (Var ());
    vals.push_back (0);
    phases.push_back (0);
    if (!new_idx) continue;
    decisions.push (new_idx);
#ifndef NLGL
    lglfreeze (lgl, new_idx);
#endif
  }
}

// All assigned literals propagated.
//
bool Solver::propagated () { return next == trail.size (); }

// Boolean constraint propagation (BCP) is one of the most expensive
// functions in SAT solvers.  We use blocking literals to reduce memory
// accesses, which also gives us faster propagation for binary clauses.
//
Clause * Solver::bcp () {
  long visits = 0, propagations = 0;
  Clause * conflict = empty;
  while (!conflict && next < trail.size ()) { // BFS over assigned literals
    propagations++;
    int lit = -trail[next++];                 // check negative occurrences
    LOG ("propagating " << -lit);
    Watches & ws = watches (lit);
    vector<Watch>::const_iterator i;
    vector<Watch>::iterator j = ws.begin ();
    for (i = j; !conflict && i != ws.end (); i++) {
      Watch w = *j++ = *i;		// by default copy watch
      int other = w.blit;		// invariant: blit occurs in clause
      Value v = val (other);		// value of blocking literal
      if (v == TRUE) continue;		// blocking literal is true
      Clause * clause = w.clause;
      if (ignore) {				// if 'ignore is non-zero
	 if (ignore == clause) continue;	// ignore this clause
	 if (clause->redundant) { 		// and all redundant clauses
	   if (!w.binary) visits++;		// count as visited clause
	   continue;
	 }
      }
      if (w.binary) {				// no need to access clause
	if (v == FALSE) conflict = clause;	// other literal false too
	else assign (other, clause);            // other literal unit
      } else {
	visits++;
	if (clause->dumped) { j--; continue; }	// drop watch if dumped
	int * lits = clause->lits, * p;
	if (lits[0] == lit) swap (lits[0], lits[1]); // normalize lit pos
	assert (lits[1] == lit);                     // => simpler code below
	for (p = lits + 2; (other = *p); p++) // search first non-false
	  if ((v = val (other)) >= 0) break;  // lit as replacement watch
	if (other) {
	  *p = lit, lits[1] = other;                 // swap old with new 
	  add_watch (other, lits[0], false, clause); // add new watch
	  j--;					     // drop old watch
	} else { // no non-false replacement found
	  v = val (other = lits[0]);                  // get other watch
	  if (v == FALSE) conflict = clause;	      // all lits false
	  else if (v != TRUE) assign (other, clause); // other watch unit
          else j[-1].blit = other;                    // optimistic update

	}
      }
    }
    // If there is a conflict we bailed out of the loop early and thus have
    // to still copy the remaining watches before resetting the sizes.
    if (conflict) while (i != ws.end ()) *j++ = *i++;
    ws.resize (j);
  }
  if (conflict) {
    LOGCLAUSE (conflict, "conflicting");
    if (!simplifier) stats.conflicts++;   // otherwise counted differently
  }
  stats.propagations += propagations;
  stats.steps += visits;
  return conflict;
}

// Mark this frame as being used to compute the LBD=glue value for
// better predicint learned clause usefullness as in Glucose.
//
bool Solver::mark_frame (int lit) {
  int level = var (lit).level;
  Frame & frame = control[level];
  if (frame.mark) return false;
  frame.mark++;
  frames.push_back (level);
  return true;
}

// Unmark all frames and return their number (glue = LBD).
//
int Solver::unmark_frames () {
  int res = (int) frames.size ();
  while (!frames.empty ()) {
    Frame & f = control[frames.back ()];
    frames.pop_back ();
    assert (f.mark);
    f.mark = 0;
  }
  return res;
}

// Rescore all scores and the score_increment to be at most 1,
// but keep the relative order of the priorities.
//
void Solver::rescore () {
  double max_score = score_increment;
  for (int idx = 1; idx < (int) vars.size (); idx++) {
    double p = decisions.priority (idx);
    if (p > max_score) max_score = p;
  }
  double factor = 1/max_score;
  LOG ("rescoring by " << factor);
  decisions.rescore (factor);
  score_increment *= factor;
}

// Bump (increases) score of a literal according to the EVSIDS, the
// exponential VSIDS scheme of MiniSAT.
//
void Solver::bump_lit (int lit) {
  const double max_priority = 1e300;
  int idx = abs (lit);
  double old_priority;
  if (score_increment > max_priority ||			// avoid overflow
      (old_priority = decisions.priority (idx)) > max_priority) {
    rescore ();
    old_priority = decisions.priority (idx);
  }
  double new_priority = old_priority + score_increment;
  decisions.update (idx, new_priority);
  LOG ("bumping score of " << idx  << " from " 
       << scientific << setprecision (3)
       << old_priority << " to " << new_priority);
}

// If this literal has already been analyzed just return.  Otherwise
// mark it as having been analyzed and bump its score.  If then the
// literal is assigned on the current level return true to tell the
// caller that another literal on the current level is marked.
// Finally if the literal has not been marked before and was
// assigned on a smaller decision level add it to the learned clause.
//
bool Solver::pull_lit (int lit) {
  if (val (lit) == TRUE) return false;
  if (marked (lit)) return false;
  mark (lit);
  bump_lit (lit);
  if (mark_frame (lit)) LOG ("pulling frame " << var (lit).level);
  if (var (lit).level == level) return true;
  LOG ("adding literal " << lit);
  addedlits.push_back (lit);
  return false;
}

// Bump (increase) activity of a clause by simply counting number of times a
// clauses was involved in conflict analysis.

void Solver::bump_clause (Clause * c) { c->activity++; }

// Minimize the first UIP clause by trying to remove this literal.
// If this lieral can be removed true is returned and all literals needed in
// recursive resolutions to remove it are marked.
//
// This test is a simple DFS which starts at the antecedent literals of the
// given starting literal.  It goes recursively backward over reason clauses
// and is successful if DFS eventually reaches marked literals and no
// unmarked decision.
//
bool Solver::minimize_lit (int root) {
  assert (marked (root));
  Clause * reason = var (root).reason;
  if (!reason) return false;
  size_t old_seen_size = seen.size (), next_seen = old_seen_size;
  bool res = true;
  int lit = root;
  for (;;) {
    int other;
    for (const int * p = reason->lits; res && (other = *p); p++) {
      if (other == lit) continue;
      assert (val (other) == FALSE);
      if (marked (other)) continue;
      Var & v = var (other);
      if (!v.reason) res = false;
      else if (!control[v.level].mark) res = false;
      else mark (other);
    }
    if (!res || next_seen == seen.size ()) break;
    lit = -seen[next_seen++];
    reason = var (lit).reason;
    assert (reason);
  } 
  if (res) LOG ("minimizing " << lit);
  else unmark (old_seen_size);
  return res;
}

// Minimization  of the learned first UIP clause.
//
void Solver::minimize_clause () {
  LOG ("minimize 1st UIP clause");
  start (&stats.time.minimize);
  size_t learned = addedlits.size ();
  stats.lits.learned += learned;

  vector<int>::const_iterator i;
  vector<int>::iterator j = addedlits.begin ();
  for (i = j; i != addedlits.end (); i++)
    if (!minimize_lit (-*i)) *j++ = *i;
  addedlits.resize (j);

  stats.lits.minimized += learned - addedlits.size ();
  stop ();
}

// Main analysis function, which learns a first UIP clause from the
// given conflict.  It traverses in reverse assignment order reasons
// of literals assigned on the current decision level which are backward
// reachable in the implication graph from the conflict.
//
void Solver::analyze (Clause * reason) {
  if (empty) { assert (!level); return; }
  start (&stats.time.analyze);
  assert (addedlits.empty ());
  vector<int>::const_iterator it = trail.end ();
  int lit = 0, open = 0;
  for (;;) {
    LOGCLAUSE (reason, "analyzing");
    bump_clause (reason);
    for (const int * p = reason->lits; (lit = *p); p++)
      if (pull_lit (lit)) open++;
    while (it != trail.begin() && !marked (lit = -*--it))
      assert (var (lit).level == level);
    if (it == trail.begin () || !--open) break;
    reason = var (lit).reason;
    assert (reason);
  }
  assert (lit);			// TODO remove this coverage hole
  if (lit) {
    LOG ("adding UIP " << lit);
    addedlits.push_back (lit);
  }
  minimize_clause ();
  unmark ();			// flush seen and unmark all literals
  assert (clause_entailed ());
  int glue = unmark_frames ();
  connect_clause (new_clause (true, glue));
  addedlits.clear ();
  score_increment *= opts.scincfact/1000.0;
  if (!level && !empty) stats.iterations++, report ('i');
  stop ();
}

/*------------------------------------------------------------------------*/

int Solver::remaining_vars () const {
  int res = maxvar ();
  res -= stats.vars.fixed;
  res -= stats.vars.equivalent;
  res -= stats.vars.eliminated;
  return res;
}

void Solver::report (char type) {
  if (!opts.verbose) return;
  if (!(stats.reported++ % 22)) {
    print_line ();
    cout << "c    seconds        irredundant          redundant" << endl
         << "c          variables           conflicts";

    if (monitor_space) cout << "             MB";	// not thread safe!
    cout << endl;
    print_line ();
  }
  cout 
  << "c " << type << ' '
  << setprecision(1) << setw (7) << fixed << seconds () << ' '
  << setprecision(0) << setw (7) << fixed << remaining_vars () << ' '
  << setw(9) << stats.clauses.irredundant << ' '
  << setw(9) << stats.conflicts << ' '
  << setw(8) << stats.clauses.redundant;
  if (monitor_space)
    cout << setprecision(1) << setw(7) << (current_bytes/(double)(1<<20));
  cout << endl << flush;
}

/*------------------------------------------------------------------------*/

// Compute the next conflict number at which a restart is triggered.  This
// is Luby controlled (rapid restarts) with basic interval 'restartint'.

void Solver::new_restart_limit () {
  long new_interval = opts.restartint * luby (stats.restarts + 1);
  bool doreport = (stats.restarts == 1);

  if (new_interval > limits.max_restart_interval) {
    limits.max_restart_interval = new_interval;
    doreport = true;
  }

  limits.restart = stats.conflicts + new_interval;

  // Do not report every 'restart', but only those where the interval to the
  // next restart reached a new maximum (and do not miss the first one).

  if (doreport) report ('R');
}

bool Solver::restarting () const {
  if (!opts.restart) return false;
  return stats.conflicts >= limits.restart;
}

void Solver::restart () {
  start (&stats.time.restart);
  LOG ("restart");
  stats.restarts++;

  // Get the next decision by popping from the 'decisions' priority queue as
  // long the top is still assigned.  This is the work 'decide' would need
  // to do anyhoze. We have to keep the first unassigned variable on the
  // decision queue.
  //
  int next_decision = 0;
  while (!next_decision && !decisions.empty ()) {
    int lit = decisions.top ();
    if (val (lit)) decisions.pop (lit);
    else next_decision = lit;
  }

  if (next_decision) {		// Is there an unassigned variable left?

    // Find the outermost decision with smaller priority.  This decision has
    // to be backtracked over, but the rest of the trail can be reused.

    double next_decision_priority = decisions.priority (next_decision);
    int new_level;
    for (new_level = 0; new_level < level; new_level++) {
      Frame & frame = control[new_level + 1];
      int decision = abs (frame.decision);
      if (decisions.priority (decision) < next_decision_priority) {
	// At level 'new_level + 1' there is a 'decision' with smaller
	// priority than the 'next_decision' so backtrack to decision level
	// 'new_level' in order to use 'next_decision' as decision at that
	// level instead.
	break;
      }
    }

    // If all previous decisions have smaller priority than 'next_decision'
    // then 'new_level' will reach 'level' and no resstart happens.

    backtrack (new_level);
  }

  new_restart_limit ();		// update restart limit
  stop ();
}

/*------------------------------------------------------------------------*/

// Check if the number of redundant clauses without the forcing redundant
// clauses is larger than the reduce limit.
//
bool Solver::reducing () {
  if (!limits.reduce.redundant) limits.reduce.redundant = opts.redinit;
  long limit = limits.reduce.redundant;
  limit += limits.reduce.forcing;
  limit += limits.reduce.important;
  return limit <= stats.clauses.redundant;
}

void Solver::reduce () {
  start (&stats.time.reduce);
  LOG ("reduce");
  stats.reductions++;

  {
    Clause * c;

    // First store all redundant and non-forcing clauses in 'candidates'.
    //
    vector<Clause *> candidates;
    vector<Clause*>::const_iterator i;
    for (i = clauses.begin (); i != clauses.end (); i++)
      if ((c = *i)->redundant && !c->important && !c->forcing)
	candidates.push_back (c);

    size_t keep = candidates.size ()/2;		// reduce by half

    LOG ("reduction " << stats.reductions << " keeps " <<
         keep << " clauses " << " out of " << candidates.size ());

    // Now sort with respect to glue and activity.
    //
    candidates.sort (Clause::cmp);

    // Mark those that are going to be removed as garbage.
    //
    for (size_t i = keep; i < candidates.size (); i++)
      candidates[i]->remove = true;

    // Flush references to garbage clauses from watches.
    //
    for (int idx = 1; idx <= maxvar (); idx++)
      for (int sign = -1; sign <= 1 ;sign += 2) {
	Watches & ws = watches (sign * idx);
	vector<Watch>::const_iterator i;
	vector<Watch>::iterator j = ws.begin ();
	for (i = j; i != ws.end (); i++)
	  if (!(*i).clause->remove) *j++ = *i;
	ws.resize (j);
      }

    // Flush references to garbage clauses from 'clauses'.
    {
      size_t i, j = 0;
      for (i = 0; i < clauses.size (); i++) {
	c = clauses[i];
	if (i == distilled) distilled = j;
	if (!c->remove) clauses[j++] = c;
      }
      if (i == distilled) distilled = j;
      clauses.resize (j);
    }

    // Finally delete those removed clauses.
    //
    long reduced = 0;
    for (size_t i = keep; i < candidates.size (); i++)
      delete_clause (candidates[i]), reduced++;
    stats.clauses.reduced += reduced;
    LOG ("reduced " << reduced << " clauses");
  }		

  // The reduce limit is increased in a simple arithmetic way, which results
  // in the conflict number at which reduction occurs to form a second order
  // arithmetic series.

  if (!limits.reduce.increment) limits.reduce.increment = opts.redinc;
  limits.reduce.redundant += limits.reduce.increment;
  limits.reduce.increment += opts.redinc;

  report ('-');
  stop ();
}

/*------------------------------------------------------------------------*/

void Solver::assume (int decision) {
  assert (propagated ());
  level++;
  int height = (int) trail.size ();
  LOG ("push frame decision " << decision << " trail " << height);
  control.push_back (Frame (decision, level, height));
  assert (level + 1 == (int) control.size ());
  assign (decision);
}

// Check if there are unassigned literals left.  If all are assigned return
// false.  Otherwise return the one with maximium score in the decisions
// priority dequeue.
//
bool Solver::decide () {
  assert (propagated ());
  int decision = 0;
  while (!decision && !decisions.empty ()) {
    int lit = decisions.top ();
    decisions.pop (lit);
    if (!val (lit)) decision = lit;
  }
  if (!decision) return false;
  assert (decision > 0);
  if (phases[decision] < 0) decision = -decision;
  stats.decisions++;
  assume (decision);
  return true;
}

/*------------------------------------------------------------------------*/

int Solver::size_penalty () const {
  long num_clauses = (stats.clauses.irredundant) >> 17;
  int logres = 0;
  while (num_clauses && logres < 5) num_clauses >>= 1, logres++; 
  return 1 << logres;
}

void Solver::distill (long steps) {
  assert (!level);
  if (!opts.distill) return;
  start (&stats.time.distill);
  LOG ("distill");

  long limit = stats.steps + steps / size_penalty ();
  bool changed = false, done = clauses.empty ();
  if (distilled >= clauses.size ()) distilled = 0;

  while (!empty && !done && stats.steps < limit) {

    Clause * c = clauses[distilled];

    if (!c->redundant &&		// distill only irredundant clauses
        !c->satisfied &&
        !c->dumped ) {

      // First get real size ignoring false literals.
      // Also check clause for being top-level satisfied.
      //
      int lit, size = 0;
      for (const int * p = c->lits; (lit = *p); p++) {
	Value v = val (lit);
	if (v == FALSE) continue;
	if (v == TRUE) { c->satisfied = true; break; }
	size++;
      }

      // Do not distill satisfied nor binary clauses.
      //
      if (!c->satisfied && size > 2) {
	Clause * conflict = 0;
	assert (!ignore);
	ignore = c;	// BCP without this clause, ignore redundant clauses
	for (const int * p = c->lits; !conflict && (lit = *p); p++) {
	  Value v = val (lit);
	  if (v == FALSE) continue;
	  if (v == TRUE) break;		// TODO strengthen then set 'changed'
	  assume (-lit);
	  conflict = bcp ();
	}
	ignore = 0;
	if (conflict) {
	  analyze (conflict);
	  if (empty) LOG ("distilled empty clause");
	  else if (!level) {
	    assert (next < trail.size ());
	    LOGCLAUSE (c, "unit by distilling");
	    stats.distill.units++;
	    changed = true;
	    conflict = bcp ();
	    if (conflict) analyze (conflict), assert (empty);
	  } else {
	    LOGCLAUSE (c, "subsumption by distilling");
	    stats.distill.subsumed++;
	    dump_clause (c);
	  }
	}
	backtrack ();
      }
    }
    if (++distilled == clauses.size ()) {
      if (changed) changed = false;		// continue
      else done = true;				// one round without change
      distilled = 0;
    }
  }
  stop ();
  report ('l');
}

/*------------------------------------------------------------------------*/

// Check clause to be satisfied by current assignment.

bool Solver::satisfied (Clause * c) const {
  int lit;
  if (c->satisfied) return true;
  for (const int * p = c->lits; (lit = *p); p++)
    if (val (lit) == TRUE) {
      if (!level) c->satisfied = true;
      return true;			// here: return early
    }
  return false;
}

/*------------------------------------------------------------------------*/

// Try to resolve two clauses.  Returns true iff resolvent is non trivial.

bool Solver::try_resolve (Clause * c, int pivot, Clause * d) {
  assert (!c->dumped && !c->satisfied);
  assert (!d->dumped && !d->satisfied);
  bool res = true;
  assert (seen.empty ());
  int lit;
  LOGCLAUSE (c, "try resolve 1st antecedent");
  stats.steps++;
  for (const int * p = c->lits; (lit = *p); p++) {
    if (lit == pivot) continue;
    assert (!marked (lit));
    mark (lit);
  }
  LOGCLAUSE (d, "try resolve 2nd antecedent");
  stats.steps++;
  for (const int * p = d->lits; res && (lit = *p); p++) {
    if (lit == -pivot) continue;
    if (marked (lit)) continue;
    if (marked (-lit)) res = false;
    else mark (lit);
  }
  unmark ();
  if (res) LOG ("resolvent would not be tautological");
  else LOG ("resolvent wouldbe tautological");
  return res;
}

/*------------------------------------------------------------------------*/

// Determine whether resolution of all positive with all negative
// occurrences of 'cand' produce at most as many non trivial resolvents as
// the number of positive plus negative occurrences.

bool Solver::try_elim (int cand) {
  assert (var (cand).free ());
  Occs & p = occs (cand), & n = occs (-cand);
  long limit = p.count + n.count;
  vector<Clause *>::const_iterator i;
  for (i = p.clauses.begin (); limit >= 0 && i != p.clauses.end (); i++) {
    Clause * c = *i;
    stats.steps++;
    if (c->dumped || satisfied (c)) continue;
    vector<Clause *>::const_iterator j;
    for (j = n.clauses.begin (); limit >= 0 && j != n.clauses.end (); j++) {
      Clause * d = *j;
      stats.steps++;
      if (d->dumped || satisfied (d)) continue;
      if (try_resolve(c, cand, d)) limit--;
    }
  }
  if (limit < 0) LOG ("elim " << cand << " failed");
  else LOG ("elim " << cand << " removes " << limit << " clauses");
  return limit >= 0;
}

/*------------------------------------------------------------------------*/

// Now really do resolve those clauses and add resolvent.

void Solver::do_resolve (Clause * c, int pivot, Clause * d) {
  assert (!c->dumped && !c->satisfied);
  assert (!d->dumped && !d->satisfied);
  assert (addedlits.empty ());
  int lit;
  LOGCLAUSE (c, "do resolve 1st antecedent");
  stats.steps++;
  for (const int * p = c->lits; (lit = *p); p++)
    if (lit != pivot) addedlits.push_back (lit);
  LOGCLAUSE (d, "do resolve 2nd antecedent");
  stats.steps++;
  for (const int * p = d->lits; (lit = *p); p++)
    if (lit != -pivot) addedlits.push_back (lit);
  if (trivial_clause ()) LOG ("resolvent tautological");
  else connect_clause (new_clause ());
  addedlits.clear ();
}

/*------------------------------------------------------------------------*/

void Solver::push_extension (int lit) {
  extension.push_back (lit);
  LOG ("push extension " << lit);
}

// Really do eliminate 'cand' by variable elimination.

void Solver::do_elim (int cand) {

  assert (schedule);
  assert (var (cand).free ());

  // all resolutions on 'cand'
  Occs & p = occs (cand), & n = occs (-cand);
  vector<Clause *>::const_iterator i;
  for (i = p.clauses.begin (); i != p.clauses.end (); i++) {
    Clause * c = *i;
    stats.steps++;
    if (c->dumped || satisfied (c)) continue;
    vector<Clause *>::const_iterator j;
    for (j = n.clauses.begin (); j != n.clauses.end (); j++) {
      Clause * d = *j;
      stats.steps++;
      if (d->dumped || satisfied (d)) continue;
      do_resolve (c, cand, d);
    }
  }

  // save clauses with smaller occurences on extension stack
  {
    int extend;
    Occs * e;
    if (p.count < n.count) extend = cand, e = &p;
    else extend = -cand, e = &n;
    for (i = e->clauses.begin (); i != e->clauses.end (); i++) {
      Clause * c = *i;
      if (c->dumped || satisfied (c)) continue;
      int lit;
      push_extension (0);
      stats.steps++;
      for (const int * p = c->lits; (lit = *p); p++)
	if (lit != extend) push_extension (lit);
      push_extension (extend);
    }
    push_extension (0);
    push_extension (-extend);
  }

  // dump clauses with positive occurrences

  while (!p.clauses.empty ()) {
    Clause * c = p.clauses.back ();
    stats.steps++;
    p.clauses.pop_back ();
    if (c->satisfied || c->dumped) continue;
    stats.clauses.eliminated++;
    dump_clause (c);
  }
  p.clauses.release ();

  // dump clauses with negative occurrences

  while (!n.clauses.empty ()) {
    Clause * c = n.clauses.back ();
    stats.steps++;
    n.clauses.pop_back ();
    if (c->satisfied || c->dumped) continue;
    stats.clauses.eliminated++;
    dump_clause (c);
  }
  n.clauses.release ();

  LOG ("eliminated " << cand);
  var (cand).state = Var::ELIMINATED;
  stats.vars.eliminated++;

  Clause * conflict = bcp ();
  if (conflict) analyze (conflict), assert (empty);
}

/*------------------------------------------------------------------------*/

bool Solver::contains_eliminated (Clause * c) const {
  int lit;
  for (const int * p = c->lits; (lit = *p); p++)
    if (var (lit).state == Var::ELIMINATED) return true;
  return false;
}

void Solver::dump_eliminated () {
  vector<Clause*>::const_iterator i;
  for (i = clauses.begin (); i != clauses.end (); i++) {
    Clause * c = *i;
    if (!c->redundant || c->satisfied || c->dumped) continue;
    if (contains_eliminated (c)) dump_clause (c);
  }
}

/*------------------------------------------------------------------------*/

// Bounded Variable Elimination = Clause Distribution = DP 
// This is the same bounded variable elimination technique as in SatELite.

void Solver::elim (long steps) {
  assert (!level);
  if (!opts.elim) return;
  start (&stats.time.elim);
  if (!dense) connect_occs ();
  assert (simplifier == TOPSIMP), simplifier = ELIM;
  assert (!schedule);
  schedule = cands.elim.empty ();
  for (int idx = 1; idx <= maxvar (); idx++) touch (idx);
  long limit = stats.steps + 10*steps/size_penalty ();
  schedule = true;
  vector<Clause*>::const_iterator i;
  for (i = clauses.begin (); i != clauses.end (); i++) {
    Clause *c = *i;
    if (satisfied (c)) for (const int *p = c->lits; *p; p++) touch (*p);
  }
  while (!empty && !cands.elim.empty () && stats.steps < limit) {
    int cand = cands.elim.top ();
    long priority = cands.elim.priority (cand);
    cands.elim.pop (cand);
    if (!priority || !var (cand).free ()) continue;
    LOG ("next elim cand " << cand << " priority " << priority);
    if (try_elim (cand)) do_elim (cand);
  }
  assert (schedule), schedule = false;
  assert (simplifier == ELIM), simplifier = TOPSIMP;
  dump_eliminated ();
  report ('e');
  stop ();
}

/*------------------------------------------------------------------------*/

void Solver::disconnect_clauses () {

  limits.reduce.important = 0;		// see comment in 'connect_clause'

  for (int idx = 1; idx <= maxvar (); idx++)
    for (int sign = -1; sign <= 1; sign += 2) {
      int lit = sign*idx;
      watches (lit).release ();
      occs (lit).release ();
    }
  dense = false;
}

// Remove false literals from clause which has the main benefit of shorting
// clauses for faster 'bcp', in particular if the clause becomes binary.
// We simply keep the same memory to avoid case splits for 'forcing'
// clauses.  This however means we need to store the actual 'allocated'
// size in the clause which additionally wastes memory.

Clause * Solver::reduce_clause (Clause * c) {
  int * j = c->lits, lit;
  for (const int * i = j; (lit = *i); i++)
    if (val (lit) >= 0) *j++ = lit;
  *j = 0;
  return c;
}

// Collect and delete all top-level satisfied or dumped non-forcing clauses.

void Solver::collect_clauses () {
  assert (!level);			    // we are at the top-level
  size_t i, j = 0;
  long collected = 0;
  for (i = j; i < clauses.size (); i++) {
    if (i == distilled) distilled = j ;     // fix distilled
    Clause * c = clauses[i];
    if (satisfied (c)) c->satisfied = true; // satisfied at top-level
    if (c->forcing ||			    // keep forcing reason clauses and
        !(c->dumped || c->satisfied))	    // neither dumped nor satisfied
      clauses[j++] = reduce_clause (c);
    else delete_clause (c), collected++;
  }
  if (i == distilled) distilled = j;
  clauses.resize (j);
  clauses.shrink ();
  stats.clauses.collected += collected;
  LOG ("collected " << collected << " clauses");
}

void Solver::connect_clauses () {
  assert (!dense);
  vector<Clause*>::const_iterator i;
  for (i = clauses.begin (); i != clauses.end (); i++)
    connect_clause (*i);
}

void Solver::collect () {
  start (&stats.time.collect);
  LOG ("collect");
  disconnect_clauses ();
  collect_clauses ();
  connect_clauses ();
  report ('c');
  stop ();
}

/*------------------------------------------------------------------------*/

// Inprocessing simplifies the instance interleaved with CDCL search.
// The steps limit given as arguments restricts the amount of work allowed
// in each inprocessing interval.
//
Status Solver::simplify (long steps) {
  start (&stats.time.simplify);
  LOG ("simplify");
  assert (!simplifier);
  simplifier = TOPSIMP;		// disables phase saving, computes different stats
  stats.simplifications++;
  backtrack ();			// simplification restarts implicitly 
  if (!empty) distill (steps);	// distillation
  if (!empty) elim (steps);	// bounded variable elimination

  /* lots of other inprocessors missing:
   *
   * probing (tree based, simple, lifting)
   * equivalence reasoning
   * (asymmetric) blocked (covered) clause elimination
   * transitive reduction
   * congruence closure
   * unhiding
   *
   * etc.
   */

  collect ();			// collect garbage clauses

  assert (simplifier = TOPSIMP);
  simplifier = NOSIMP;
  stop ();
  return empty ? UNSATISFIABLE : UNKNOWN;
}

// This is the CDCL search loop.  It is limited by the number of conflicts,
// since we interleave it with simplification through inprocessing.
//
Status Solver::search (long limit) {
  start (&stats.time.search);
  LOG ("search");
  long conflicts = 0;
  Clause * conflict;
  Status res = UNKNOWN;
  while (!res)
    if (empty) res = UNSATISFIABLE;
    else if ((conflict = bcp ())) analyze (conflict), conflicts++;
    else if (conflicts >= limit) break;
    else if (reducing ()) reduce ();
    else if (restarting ()) restart ();
    else if (!decide ()) res = SATISFIABLE;
  stop ();
  return res;
}

/*------------------------------------------------------------------------*/

void Solver::print () const {
  cout << "p cnf " << maxvar () << ' ' << clauses.size () << endl;
  vector<Clause*>::const_iterator i;
  for (i = clauses.begin (); i != clauses.end (); i++)
    (*i)->print ();
}

/*------------------------------------------------------------------------*/

// Extend current partial assignment to full assignment using the clauses
// saved on the extension stack.  At the same time flip assigned literals
// which would falsify saved clauses.  The extension stack is filled in such
// a way that this will always produce a correct satisfying assignment for
// the original clauses.  This techniques was used first in MiniSAT by
// Niklas Soerensson and is also described in our SAT'10 resp. IJCAR'12
// papers.
//
void Solver::extend () {
  start (&stats.time.extend);
  LOG ("extend");
  while (!extension.empty ()) {
    int lit = extension.back (), other;
    bool satisfied = false;
    while ((other = extension.back ())) {
      extension.pop_back ();
      if (val (other) == TRUE) satisfied = true;
    }
    extension.pop_back ();
    if (!satisfied) {
      LOG ("flipping assignment of " << lit);
      vals[abs (lit)] = sign (lit);
    }
  }
  stop ();
}

// Return TRUE if one of the literals traversed is TRUE.  Continue
// traversing and advancing the iterator until a zero sentinel is found.

bool Solver::satisfied (vector<int>::const_iterator & it) const {
  bool res;
  int lit;
  for (res = false; (lit = *it); it++)
    if (val (lit) == TRUE) res = true;	// do not return!
  return res;
}

void Solver::check_satisfied () const {
  vector<int>::const_iterator c;
  for (c = original.begin (); c != original.end (); c++)
    if (!satisfied (c))
      internal_error ("invalid assignment");
}

/*------------------------------------------------------------------------*/
/**** PART6: implemenetation of public member functions *******************/
/*------------------------------------------------------------------------*/

Solver::Solver () : 
  level (0),			// top-level = root-level = 0
  dense (false),		// start in sparse mode only watching clauses
  simplifier (NOSIMP),		// only non zero within 'simplify'
  schedule (false),		// do not schedule touched literals
  next (0),			// nothing to propagate yet
  empty (0),			// no empty clause found yet
  score_increment (1),		// initial bump score increment
  ignore (0),			// use all clauses during bcp
  distilled (0)			// all before clauses[distilled] distilled
{ 
  set_fpu_to_double_precision ();
  inc_bytes (sizeof *this);
  control.push_back (Frame ());	// empty frame for top-level without decision
#ifndef NLGL
  lgl = lglinit ();
#endif
}

/*------------------------------------------------------------------------*/

Solver::~Solver () {

  // Beside memory for the vectors in containers declared as members of
  // 'Solver' the only other memory we allocate is for clauses and for the
  // watch and occurrence list containers of variables.

  while (!clauses.empty ()) {
    Clause * c = clauses.back ();
    clauses.pop_back ();
    delete_clause (c);
  }

  while (!vars.empty ()) {	// 'vector' does not call destructors
    Var & v = vars.back ();
    v.release ();		// so need to release variables explicitly
    vars.pop_back ();
  }

#ifndef NLGL
  lglrelease (lgl);
#endif

  if (monitor_space) dec_bytes (sizeof *this);	// not thread safe
}

/*------------------------------------------------------------------------*/

void Solver::addlit (int lit) {
  original.push_back (lit);
  if (lit) {
    import (lit);
    addedlits.push_back (lit);
  } else {
    if (!trivial_clause ()) connect_clause (new_clause ());
    addedlits.clear ();
  }
}

/*------------------------------------------------------------------------*/

Status Solver::solve () {

  if (!addedlits.empty ()) internal_error ("API usage: zero literal missing");

  start (&stats.time.solve);
  LOG ("solve");

  // Set up limits.
  //
  long steps = opts.simpint;	// limit steps in simplification round
  long conflicts = 0;	    	// limit conflicts in search round
  new_restart_limit ();	    	// intialize restart limit (search)
  bias_phases ();	    	// compute initial phase bias

  Status res;

  //------------------------------------------------------------------------
  // Main Solving Loop
  //
  // This is the main solving loop, which consists of limited interleaved
  // CDCL search and simplifications through inprocessing.
  //
  for (;;)
    if ((res = search (conflicts))) break;
    else if ((res = simplify (steps))) break;
    else conflicts += opts.searchint, steps += opts.simpint;
  //------------------------------------------------------------------------

  char type;				// for reporting result
  switch (res) {
    default:
      type = '?';
      break;
    case SATISFIABLE:
      type = '1'; 
      extend ();			// extend partial assignemtn to full
      check_satisfied ();		// check assignment
      break;
    case UNSATISFIABLE: 
      type = '0';
#ifndef NLGL
      if (opts.check) assert (lglsat (lgl) == 20);  // check consistency
#endif
      break;
  }

  report (type);
  if (opts.verbose && stats.reported) print_line ();
  stop ();

  return res;
}

/*------------------------------------------------------------------------*/

Value Solver::value (int lit) { import (lit); return val (lit); }

/*------------------------------------------------------------------------*/

int Solver::option (const char * str) {		// get current option value
#undef OPTION
#define OPTION(NAME,DEFAULT,MIN,MAX,DESCRIPTION) \
  if (!strcmp (# NAME, str)) return opts.NAME.val;
#include "options.h"
  return 0;
}

void Solver::option (const char * str, int val) {	// set option value
#undef OPTION
#define OPTION(NAME,DEFAULT,MIN,MAX,DESCRIPTION) \
  if (!strcmp (# NAME, str)) { \
    assert (opts.NAME.min <= opts.NAME.max); \
    if (val > opts.NAME.max) val = opts.NAME.max; \
    if (val < opts.NAME.min) val = opts.NAME.min; \
    opts.NAME.val = val; \
    return; \
  }
#include "options.h"
}

/*------------------------------------------------------------------------*/

// Print the time of one solving phase (or subphase) relative to the total
// time used for solving.  We keep it this part even though it is internal.

void Solver::print_time_for (double time, const char * name) {
  cout 
  << "c "
  << fixed << setprecision (2) << setw (8) << right << time
  << ' ' << setw (10) << left << name
  << ' ' << fixed << setprecision (0) << right << setw (3) 
  << percent (time,stats.time.solve) << '%'
  << endl;
}

#define PRINT_TIME_FOR(NAME) print_time_for (stats.time.NAME, #NAME)

void Solver::statistics () {
  print_line ();
  cout << "c sizeof (Var) = " << sizeof (Var) << endl;
  cout << "c sizeof (Occs) = " << sizeof (Occs) << endl;
  cout << "c sizeof (Watch) = " << sizeof (Watch) << endl;
  cout << "c sizeof (Watches) = " << sizeof (Watches) << endl;
  cout << "c sizeof (Clause) = " << Clause::bytes (0)
       << " + size * " << sizeof (int) << endl;
  print_line ();
  while (!timers.empty ()) stop ();
  double solve = stats.time.solve;
  cout
   << fixed << setprecision (0)
  << "c " 
  << stats.iterations << " iterations, "
  << stats.restarts << " restarts" << endl
  << "c " 
  << stats.simplifications << " simplifications, "
  << stats.reductions << " reductions" << endl
  << "c " 
  << stats.decisions << " decisions, "
  << average (stats.decisions, solve) << " decisions per second" << endl
  << "c "
  << stats.conflicts << " conflicts, " 
  << average (stats.conflicts, solve) << " conflicts per second" << endl
  << "c " 
  << stats.lits.learned << " learned literals, "
  << stats.lits.minimized << " minimized literals " 
  << percent (stats.lits.minimized, stats.lits.learned) << "%" << endl
  << "c "
  << stats.propagations << " propagations, " << setprecision (2)
  << average (stats.propagations/1e6, solve) 
  << " million propagations per second" << endl;
  print_line ();
  cout
  << "c "
  << stats.distill.units << " distilled units, "
  << stats.distill.subsumed << " subsumed, "
  << stats.distill.strengthened << " strengthened" << endl
  << "c "
  << stats.vars.eliminated << " eliminated variables, "
  << stats.clauses.eliminated << " eliminated clauses" << endl
  ;
  print_line ('=');
  PRINT_TIME_FOR (analyze);
  PRINT_TIME_FOR (bias);
  PRINT_TIME_FOR (distill);
  PRINT_TIME_FOR (extend);
  PRINT_TIME_FOR (reduce);
  PRINT_TIME_FOR (restart);
  PRINT_TIME_FOR (minimize);
  print_line ();
  PRINT_TIME_FOR (search);
  print_line ('=');
  PRINT_TIME_FOR (collect);
  PRINT_TIME_FOR (distill);
  PRINT_TIME_FOR (elim);
  print_line ();
  PRINT_TIME_FOR (simplify);
  print_line ('#');
  PRINT_TIME_FOR (solve);
  cout << flush;
}

/*------------------------------------------------------------------------*/

// In order to hide solver details (and potentially allow multiple
// implementations) we use an 'AbstractSolver' interface.  This also allows
// to add all the code in this one file and leave it out of the header file.
//
AbstractSolver * Cleaneling::new_solver () { return new Solver (); }
